package Vulture::ResponseHandler;

use strict;
use warnings;

use Apache2::Const -compile => qw(OK REDIRECT FORBIDDEN);
use Apache2::RequestRec ();
use Apache2::RequestIO ();
use HTTP::Request;
use Apache2::RequestUtil();
use APR::Table;
use Net::LDAP;
use MIME::Base64;
use CGI qw/:standard/;
use Digest::MD5 qw(md5_hex md5_base64);
use Digest::SHA1  qw(sha1_hex sha1_base64);
use Crypt::CBC;
use Apache::SSLLookup;

use Vulture qw($db_uri $auth_url $proxy_name $proto &get_app &session &get_cookie $path);

sub auth_ssl {
	my ($id_auth, $dbh, $session) = @_;

	my $sth = $dbh->prepare("SELECT is_acl FROM ca WHERE id='".$id_auth."'");
	$sth->execute();
	if ((my ($is_acl) = $sth->fetchrow)) {		
		if ( $is_acl ) {
			$sth = $dbh->prepare("SELECT user_id FROM acl, auth WHERE auth.type=2 AND auth.id_auth='".$id_auth
					     ."' AND acl.auth_id=auth.id AND acl.user_id='".$session->{user}."'");
			$sth->execute;
			if (( $sth->fetchrow )) {
				return 1;
			}
			return -1;
		}
		return 1;
	}
}

sub auth_ldap {
	my ($id_auth, $dbh, $session) = @_;


	my $sth = $dbh->prepare("SELECT ldap_addr, ldap_port, ldap_dn, ldap_password, ldap_base_dn, ".
				"ldap_user_attr, ldap_group_attr, ldap_group_member, is_acl, ldap_group_is_dn ".
				"FROM ldap ".
				"WHERE id='".$id_auth."'");
	$sth->execute;
	my ($ldap_server, $ldap_port, $ldap_bind_dn, $ldap_bind_password, $ldap_base_dn, 
	    $ldap_uid_attr, $ldap_group_attr, $ldap_group_member, $is_acl, $ldap_group_is_dn) = $sth->fetchrow;
	my $ldap = Net::LDAP->new($ldap_server, port => $ldap_port);
	$ldap->bind($ldap_bind_dn, password => $ldap_bind_password);
	my $mesg = $ldap->search(base   => $ldap_base_dn,
				 filter => "(" . $ldap_uid_attr . "=" . $session->{user} . ")"
				);
	$mesg->code and return 0;
	my $object = $mesg->entry(0) or return 0;
	my $dnuser = $object->dn;
	my $result = $ldap->bind($dnuser, password => $session->{pass});
	if (!$result->code) {
		if ($is_acl) {
			$sth = $dbh->prepare("SELECT count(*) FROM acl WHERE user_id='".$session->{user}."' AND auth_id='".$id_auth."'");
			$sth->execute;
			if (($sth->fetchrow)) {
				return 1;
			}
			my $sql = "SELECT group_id FROM acl WHERE group_id != '' AND auth_id='".$id_auth."'";
			$sth = $dbh->prepare($sql);
			$sth->execute;
			if (!($sth->fetchrow)) {
				return 0;
			}
			my $filter_groups;
			while (my ($group) = $sth->fetchrow) {
				$filter_groups .= "(".$ldap_group_attr."=".$group.")";
			}

			my $filter_member;
			if ($ldap_group_is_dn) { # Active Directory
				$filter_member = $ldap_group_member . "=" . $dnuser;
			} else {
				$filter_member = $ldap_uid_attr . "=" . $session->{user};
			}
			my $mesg = $ldap->search(base   => $ldap_base_dn,				     
						 filter => "(&(|(objectclass=posixGroup)(objectclass=group))(|".$filter_groups.")(".$filter_member."))"
						);
			$mesg->code and return 0;
			my $object = $mesg->entry(0);
			$ldap->unbind;

			if ($object and $object->dn) {
				return 1;
			}
			return 0;
		}
		return 1;
	}

}

sub auth_sql {
	my ($id_auth, $dbh, $session) = @_;

	my $sth = $dbh->prepare("SELECT uri, db_user, db_pass, table_name, login_column, password_column, is_acl, password_algo FROM sql WHERE id='$id_auth'");
	$sth->execute();
	my ($uri, $db_user, $db_pass, $table_name, $login_column, $password_column, $is_acl, $password_algo, $salt) = $sth->fetchrow;
	my $cnx = DBI->connect($uri, $db_user, $db_pass);

	my $count;
	my $sql = "SELECT count(*) FROM ".$table_name." WHERE ".$login_column."='".$session->{user}."' AND ".$password_column."='";
	if ($password_algo eq "plain") {
		$sth = $cnx->prepare($sql . $session->{pass}."'");
		$sth->execute;
		$count = ($sth->fetchrow);
	} elsif ($password_algo eq "sha1") {
		$sth = $cnx->prepare($sql . sha1_hex($session->{pass})."'");
		$sth->execute;
		$count = ($sth->fetchrow);
		if (!$count) {
			$sth = $cnx->prepare($sql . sha1_base64($session->{pass})."'");
			$sth->execute;
			$count = ($sth->fetchrow);
		}
		if (!$count) {
			$sth = $cnx->prepare($sql . "='{SHA}".sha1_base64($session->{pass})."='");
			$sth->execute;
			$count = ($sth->fetchrow);
		}
	} elsif ($password_algo eq "md5") {
		$sth = $cnx->prepare($sql . md5_hex($session->{pass})."'");
		$sth->execute;
		$count = ($sth->fetchrow);
		if (!$count) {
			$sth = $cnx->prepare($sql . md5_base64($session->{pass})."'");
			$sth->execute;
			$count = ($sth->fetchrow);
		}
	} elsif ($password_algo eq "crypt") {
		$sth = $cnx->prepare("SELECT $password_column FROM ".$table_name." WHERE ".$login_column."='".$session->{user}."'");
		$sth->execute;
		while ((my ($crypted) = $sth->fetchrow)) {
			my $test = crypt($session->{pass}, substr($crypted,0, 2));
			$sth = $cnx->prepare($sql . $test."'");
			$sth->execute;
			$count = ($sth->fetchrow);
			last if $count;
		}
	}
	if ($count) {
		if ($is_acl) {
			$sth = $dbh->prepare("SELECT count(*) FROM acl WHERE pass='1' AND user_id='".$session->{user}."' AND auth_id='".$id_auth."'");
			$sth->execute;
			if (($sth->fetchrow)) {
				return 1;
			}
			return 0;
		}
		return 1;
	}
}

sub LDAP { 1 };
sub SSL  { 2 };
sub SQL  { 3 };

sub auth {
	my ($app, $dbh, $session) = @_;

	my $sth = $dbh->prepare("SELECT id_auth, type FROM auth WHERE id='".$app->{auth}."'");
	$sth->execute();
	my ($id_auth, $type) = $sth->fetchrow;

	if ($type == SSL) {
		return auth_ssl($id_auth, $dbh, $session);
	}
	return 0 if (!$session->{user} or !$session->{pass});

	if ($type == SQL) {
		return auth_sql($id_auth, $dbh, $session);
	}
	if ($type == LDAP) {
		return auth_ldap($id_auth, $dbh, $session);
	}
	return 0;
}

sub NEED_PIN { -1 };

sub post {
	my ($id_app, $dbh, $user, $pass, $pin) = @_;

	my $sth = $dbh->prepare("SELECT sso_forward_type, autologon FROM app WHERE id='".$id_app."'");
	$sth->execute;
	my ($sso_forward_type, $autologon) = $sth->fetchrow;
	if ($autologon ne "t") {
		my $query = "SELECT count(*) FROM vulture_pin WHERE id_app='".$id_app."' AND user='".$user."' AND vulture_pin='".md5_hex($pin)."'";
		$sth = $dbh->prepare($query);
		$sth->execute();
		if (!$sth->fetchrow) {
			my $query = "SELECT count(*) FROM vulture_pin WHERE id_app='".$id_app."' AND user='".$user."'";
			my $sth = $dbh->prepare($query);
			$sth->execute();
			if (!$sth->fetchrow) {
				return (0, undef);
			}
			# Password wrong (maybe changed if not cert)
			return (NEED_PIN, undef);
		}
	}
	if ($sso_forward_type eq "basic") {
		if ($autologon eq "t") {
			return (1, "Basic " . encode_base64($user . ':' . $pass));
		}
		my $sql = "SELECT value FROM profile WHERE id_app='".$id_app."' AND user='".$user."' AND field_var='vulture_login'";
		$sth = $dbh->prepare($sql);
		$sth->execute;
		my ($vulture_login) = $sth->fetchrow;
		$sql = "SELECT value FROM profile WHERE id_app='".$id_app."' AND user='".$user."' AND field_var='vulture_password'";
		$sth = $dbh->prepare($sql);
		$sth->execute;
		my ($vulture_password) = $sth->fetchrow;
		if ($pin) {
			my $cipher = Crypt::CBC->new( -key    => $pin,
						      -cipher => 'Blowfish',
						      -salt   => 1,
						    );
			$vulture_password = $cipher->decrypt(decode_base64($vulture_password));
			if ($pass and $pass ne $pin) {
				my $new_cipher = Crypt::CBC->new( -key    => $pass,
								  -cipher => 'Blowfish',
								  -salt   => 1,
								);
				$sql = "UPDATE profile SET value='".encode_base64($new_cipher->encrypt($vulture_password),"").
				  "' WHERE id_app='".$id_app."' AND user='".$user."' AND field_var='vulture_password'";
				$sth = $dbh->prepare($sql);
				$sth->execute;
				$sql = "UPDATE vulture_pin SET vulture_pin='".md5_hex($pass)."' WHERE id_app='".$id_app."' AND user='".$user."'";
				$sth = $dbh->prepare($sql);
				$sth->execute;

			}
		}
		return (1, "Basic " . encode_base64($vulture_login . ':' . $vulture_password));
	} else {
		my $post;
		if ($autologon eq "t") {
			$sth = $dbh->prepare("SELECT field_var, field_type, field_value FROM post WHERE id_app='".$id_app."'");
			$sth->execute;
			while (my ($var,$type,$value) = $sth->fetchrow) {
				if ($type eq "autologon_user") {
					$value = $user;
				}
				if ($type eq "autologon_password") {
					$value = $pass;
				}
				$post .= $var."=".$value."&";
			}
			return (1, $post);
		}
		my $sql = "SELECT post.field_type, profile.field_var, profile.value, profile.id FROM post, profile ".
		  "WHERE post.id_app=profile.id_app AND profile.field_var=post.field_var AND profile.id_app='".$id_app.
		    "' AND profile.field_var=post.field_var AND profile.user='".$user."'";
		my $sth = $dbh->prepare($sql);
		$sth->execute;
		while (my ($type, $var, $val) = $sth->fetchrow) {
			if ($type eq "password") {
				my $cipher = Crypt::CBC->new( -key    => $pin,
							      -cipher => 'Blowfish',
							      -salt   => 1,
							    );
				$val = $cipher->decrypt(decode_base64($val));

				if ($pass and $pass ne $pin) {
					my $new_cipher = Crypt::CBC->new( -key    => $pass,
									  -cipher => 'Blowfish',
									  -salt   => 1,
									);
					$sql = "UPDATE profile SET value='".encode_base64($new_cipher->encrypt($val),"")."' WHERE id_app='".$id_app.
					  "' AND user='".$user."' AND field_var='".$var."'";
					$sth = $dbh->prepare($sql);
					$sth->execute;
					$sql = "UPDATE vulture_pin SET vulture_pin='".md5_hex($pass)."' WHERE id_app='".$id_app."' AND user='".$user."'";
					$sth = $dbh->prepare($sql);
					$sth->execute;
				}
			}
			$post .= $var."=".$val."&";
		}
		return (1, $post);
	}
}

sub INVALID_PIN { -2 };

sub register {
	my ($id_app, $dbh, $user, $pass, $pin, $pin2) = @_;

	if ($pin ne $pin2) {
		return (INVALID_PIN, undef);
	}
	my $sth = $dbh->prepare("SELECT sso_forward_type, autologon FROM app WHERE id='".$id_app."'");
	$sth->execute;
	my ($sso_forward_type, $autologon) = $sth->fetchrow;
	if ($sso_forward_type eq "basic" and $autologon ne "t") {
		$sth = $dbh->prepare("INSERT INTO profile (id_app, user, field_var, value) VALUES('$id_app', '$user', 'vulture_login', '".
				     param('vulture_login')."')");
		$sth->execute;
		my $vulture_password = param('vulture_password');
		if ($pin) {
			my $cipher = Crypt::CBC->new( -key    => $pin,
						      -cipher => 'Blowfish',
						      -salt   => 1,
						    );
			$vulture_password = encode_base64($cipher->encrypt(param('vulture_password')));
		}
		$sth = $dbh->prepare("INSERT INTO profile (id_app, user, field_var, value) VALUES('$id_app', '$user', 'vulture_password', '".
				     $vulture_password."')");
		$sth->execute;
		my $val = md5_hex($pin);
		my $insert = $dbh->prepare("INSERT INTO vulture_pin (id_app, user, vulture_pin) VALUES('$id_app', '$user', '".$val."')");
		$insert->execute;
	} elsif ($autologon ne "t") {
		$sth = $dbh->prepare("SELECT field_type, field_var FROM post WHERE id_app='".$id_app."'");
		$sth->execute;
		while (my ($type, $var) = $sth->fetchrow) {
			my $val = param($var);
			if ($type eq "password") {
				my $cipher = Crypt::CBC->new( -key    => $pin,
							      -cipher => 'Blowfish',
							      -salt   => 1,
							    );
				$val = encode_base64($cipher->encrypt($val));
			}
			my $insert = $dbh->prepare("INSERT INTO profile (id_app, user, field_var, value) VALUES('$id_app', '$user', '".$var."', '".$val."')");
			$insert->execute;
		}
		my $val = md5_hex($pin);
		my $insert = $dbh->prepare("INSERT INTO vulture_pin (id_app, user, vulture_pin) VALUES('$id_app', '$user', '".$val."')");
		$insert->execute;
	}
	return post($id_app, $dbh, $user, $pass, $pin);
}

sub get_redirect {
	my ($url) = @_;

	my $redirect = <<EOF
<html>
<head>
<meta http-equiv="refresh" content="0;url=$url">
</head>
</html>
EOF
;
}

sub auth_form {
	my ($app, $dbh, $id) = @_;

	open(FH, "< $path/".$app->{id}.".tpl")
	  or open(FH, "< $path/default.tpl");
	my $html;
	while (<FH>) {
		$html .= $_;
	}
	close(FH);
	my $form = <<EOF
<form method="POST" action="/">
<input type=hidden name=vulture_app value="$id">
<table>
<tr><td>Utilisateur</td><td><input type="text" name="vulture_login"></td></tr>
<tr><td>Mot de passe</td><td><input type="password" name="vulture_password"></td></tr>
EOF
;
	$html =~ s/__FORM__/$form/xm;
	$html =~ s/__APP_NAME__/$app->{name}/xm;
	return $html;
}


sub app_form {
	my ($id_app, $dbh, $id, $pass) = @_;

	open(FH, "< $path/profile_$id_app.tpl")
	  or open(FH, "< $path/profile.tpl");
	my $html;
	while (<FH>) {
		$html .= $_;
	}
	close(FH);
	my $sth = $dbh->prepare("SELECT name, sso_forward_type FROM app WHERE id='".$id_app."'");
	$sth->execute;
	my ($name, $sso_forward_type) = $sth->fetchrow;

	my $form;
	$form = <<EOF
<table>
<form method="POST" name='app_form' action="/">
<input type=hidden name=vulture_app value=$id>
<input type=hidden name=vulture_profile value=1>
EOF
;
	if ($sso_forward_type eq "basic") {
		$form .= <<EOF
<tr><td>Utilisateur</td><td><input type="text" name="vulture_login"></td></tr>
<tr><td>Mot de passe</td><td><input type="password" name="vulture_password"></td></tr>
EOF
;
	}
	elsif ($sso_forward_type eq "post") {
		my $sth = $dbh->prepare("SELECT field_desc, field_var, field_value, field_type FROM post WHERE id_app='".$id_app.
					"' AND field_type != 'autologon_user' AND field_type != 'autologon_password'");
		$sth->execute;
		while (my ($desc, $var, $default, $type) = $sth->fetchrow) {
			$desc    =~ s/\\//g;
			$var     =~ s/\\//g;
			$default =~ s/\\//g;

			if ($type eq "hidden") {
				$form .= <<EOF
<input type="hidden" name="$var" value="$default">
EOF
;
			}
			else {
				$form .= <<EOF
<tr><td>$desc</td><td><input type="$type" name="$var" value="$default"></td></tr>
EOF
;
			}
		}
	}
	if (!$pass) {
		$form .= <<EOF
		  <tr><td>PIN</td><td><input type="password" name="vulture_pin"></td></tr>
		  <tr><td>PIN(2)</td><td><input type="password" name="vulture_pin2"></td></tr>
EOF
		      ;
	}
	$html =~ s/__FORM__/$form/xm;
	$html =~ s/__APP_NAME__/$name/xm;
	return $html;
}

sub pin_form {
	my ($id_app, $dbh, $id, $pass) = @_;

	my $field_name;
	if ($pass) {
		$field_name = "";
		open(FH, "< $path/change_pin_$id_app.tpl")
		  or open(FH, "< $path/change_pin_default.tpl");
	}
	else {
		$field_name = "PIN";
		open(FH, "< $path/pin_$id_app.tpl")
		  or open(FH, "< $path/pin_default.tpl");
	}
	my $html;
	while (<FH>) {
		$html .= $_;
	}

	my $form = <<EOF
<table>
<form method="POST" action="/">
<input type=hidden name=vulture_app value="$id">
<input type=hidden name=vulture_profile value=1>
<tr><td>$field_name</td><td><input type="password" name="vulture_pin"></td></tr>
<tr><td></td><td align="right"><input type="submit"></td></tr>
</form>
</table>
EOF
;
	$html =~ s/__FORM__/$form/xm;
	return $html;
}

sub handler {
	my $r =  Apache::SSLLookup->new(shift);

	return Apache2::Const::FORBIDDEN if ($r->uri =~ /favicon.ico$/);
	my $dbh = DBI->connect($db_uri);

	# Entering proxy
	if ($r->hostname eq $proxy_name) {
		my ($cookie_id) = get_cookie($r->headers_in->{Cookie}, 'vulture_proxy=([^;]*)');

		my $id = param('vulture_app');
		my %session;
		session(\%session, $id) if ($id);

		if (param('vulture_pin')) {
			$session{pin} = param('vulture_pin');
		}
		else {
			$session{pin} = $session{pass};
		}
		if (param('vulture_login')) {
			$session{user} = param('vulture_login');
		}
		if (param('vulture_password')) {
			$session{pass} = param('vulture_password');
		}
		if ($session{'SSL_CLIENT_M_SERIAL'} and  $session{'SSL_CLIENT_I_DN_CN'}) {
		    $session{user} ||= $session{'SSL_CLIENT_M_SERIAL'} . ";" . $session{'SSL_CLIENT_I_DN_CN'};
		}
		# Getting all information about the application we come from
		my $app = get_app($session{site}, $dbh, $r->dir_config('if_id')) if ($session{site});
		return Apache2::Const::FORBIDDEN if !$app;
		return Apache2::Const::OK if !$app->{auth};

		# Already have a proxy cookie
		my %proxy;
		session(\%proxy, $cookie_id);
		$r->headers_out->set('Set-Cookie' => "vulture_proxy=". $proxy{_session_id} . "; path=/");
		
		# We are authentified on the app $app
		if ($proxy{$app->{auth}}) {
			my $post;
			my $ret;
			# Application is configured for SSO forward
			if ($app->{is_post}) {
				if (param('vulture_profile')) {
					# Profile data just posted
					my $query = "SELECT count(*) FROM profile WHERE id_app='".$app->{id}."' AND user='".
					  $proxy{$app->{auth}}->{user}."'";
					
					my $sth = $dbh->prepare($query);
					$sth->execute;
					if (!($sth->fetchrow)) {
						# No profile create one
						($ret, $post) = register($app->{id}, $dbh, $proxy{$app->{auth}}->{user}, $proxy{$app->{auth}}->{pass},
									 $session{pin}, param('vulture_pin') ? param('vulture_pin2') : $session{pin});
					}
					else {
						($ret, $post) = post($app->{id}, $dbh, $proxy{$app->{auth}}->{user}, $proxy{$app->{auth}}->{pass},
								     $session{pin});
					}
				}
				else {
					($ret, $post) = post($app->{id}, $dbh, $proxy{$app->{auth}}->{user}, $proxy{$app->{auth}}->{pass}, $session{pin});
				}
			}
			# Display pin form
			if ($ret and $ret == NEED_PIN) {
				$session{post} = undef;
				$r->status(200);
				$r->content_type('text/html');		
				$r->print(pin_form($app->{id}, $dbh, $id, $session{pass}));
				return Apache2::Const::OK;
			}
			
			if (!$app->{is_post} or ($app->{is_post} and $ret and $ret != INVALID_PIN)) {
				chomp($post) if ($post);
				$post ||= 1;
				$session{post} = $post;
				
				$r->headers_out->set('Location' => $session{proto}."://". $session{site} . ":".$session{port}. $session{uri});
				$r->status(Apache2::Const::REDIRECT);
				return Apache2::Const::OK;
				
			}
			# Display profile form
			$session{post} = undef;
			$r->status(200);
			$r->content_type('text/html');
			$r->print(app_form($app->{id}, $dbh, $id, $session{pass}));
			return Apache2::Const::OK;
		}
		# No cookie

		my $auth = auth($app, $dbh, \%session);

		return Apache2::Const::FORBIDDEN if ($auth == -1);
		# Authentication success

		if ($auth) {
			$proxy{$app->{auth}} = {
						'user' => $session{user},
						'pass' => $session{pass},
						'pin' => $session{pin},
					       };
			my $url = $session{proto}."://".$session{site}.":".$session{port} . $session{uri};
			$r->content_type('text/html');
			$r->print(get_redirect($url));
			$r->status(200);
			return Apache2::Const::OK;
		}
		else {
			$r->status(200);
			$r->content_type('text/html');
			$r->print(auth_form($app, $dbh, $id));
			return Apache2::Const::OK;
		}
	}

	my $app = get_app($r->hostname, $dbh, $r->dir_config('if_id'));

	if ($app) {
		if (!$app->{up}) {
			open(FH, "< $path/".$app->{id}.".unavailable.tpl")
			or open(FH, "< $path/unavailable.tpl");
			my $html;
			while (<FH>) {
				$html .= $_;
			}
			close(FH);
			$html =~ s/__APP_NAME__/$r->hostname/xm;
			$r->status(200);
			$r->content_type('text/html');
			$r->print($html);
			return Apache2::Const::OK;
		}

		my %session;
		my ($cookie_id) = get_cookie($r->headers_in->{Cookie}, 'vulture_app=([^;]*)');
		session(\%session, $cookie_id);
		if ($session{post}) {
			$r->headers_out->set('Location' => $proto."://". $r->hostname .":".$r->get_server_port ."/");
			$r->status(Apache2::Const::REDIRECT);
			return Apache2::Const::OK;
		}

		$session{site} = $r->hostname;
		$session{port} = $r->get_server_port;
		$session{proto} = $proto;
		$session{uri} = $r->uri;

		$session{SSL_CLIENT_M_SERIAL} = $r->ssl_lookup('SSL_CLIENT_M_SERIAL');
		$session{SSL_CLIENT_I_DN_CN} = $r->ssl_lookup('SSL_CLIENT_I_DN_CN');
		
		my $url = $auth_url . "/?vulture_app=".  $session{_session_id};
		$r->headers_out->set('Set-Cookie' => "vulture_app=".  $session{_session_id} . "; path=/");
		$r->status(200);
		$r->content_type('text/html');
		$r->print(get_redirect($url));
		return Apache2::Const::OK;
	}
	return Apache2::Const::FORBIDDEN;
}

1;
