package Vulture::TransHandler;

use strict;
use warnings;

use Apache2::Const -compile => qw(OK REDIRECT);
use Apache2::RequestRec ();
use Apache2::RequestIO ();
use APR::Table;  
use HTTP::Request;
use Apache2::RequestUtil();
use LWP::UserAgent;
use Apache2::Connection;
use Apache::SSLLookup;
use Vulture qw(&get_app $proto &read_conf $db_uri &get_cookie &session);


sub REMOTE_ADDR { 1 };

my %headers_vars = (
		    2 => 'SSL_CLIENT_I_DN',
		    3 => 'SSL_CLIENT_M_SERIAL',
		    4 => 'SSL_CLIENT_S_DN',
		    5 => 'SSL_CLIENT_V_START',
		    6 => 'SSL_CLIENT_V_END'
		   );


sub rewrite {
	my ($r, $line, $uri) = @_;

	my ($exp, $rule) = split (/=>/, $line);
	$rule =~ s/\s*//g;
	$exp =~ s/\s*//g;

	my ($cond) = ($rule =~ /\[(.*)\]/);
	return unless ($rule and $exp and $$uri =~ /$exp/x);
	return if ($cond =~ /,NOCERT$/ and $r->ssl_lookup('SSL_CLIENT_M_SERIAL'));
	
	$rule =~ s/\[$cond\]//;
	$cond =~ s/,NOCERT$//;

	if ($cond =~ /([0-9]+)/) {
		$r->status($1);
		return $1;
	}

	if ($$uri =~ /$exp/x) { # Load $1 $2
		my @tab = split ('\$[0-9]+', $rule);
		my $i = 1;
		my $tmp;
		foreach (@tab) {
			$tmp.="'".$_."'";
			last if ($i == $#tab + 1);
			$tmp.='.$'.$i.'.';
			$i++;
		}
		$rule = eval($tmp);
	}
	$$uri =~ s|$exp|$rule|x;
	return $cond;
}


sub handler {
	my $r =  Apache::SSLLookup->new(shift);

	my $uri = $r->uri;
	my $server;
	my $dbh = DBI->connect($db_uri); 
	my $app = get_app($r->hostname, $dbh, $r->dir_config('if_id'));

	if ($app and $app->{up}) {
		my ($id) = get_cookie($r->headers_in->{Cookie}, 'vulture_app=([^;]*)');
		my %session;
		session(\%session, $id) if ($id);
		if ($session{post} or !$app->{'auth'} ) {
			if ($session{posted}
			    or !$app->{'auth'}
			    or ($session{post}
				and $app->{'sso_forward_type'} eq "basic")) {

				if (my $proxy_rules = $app->{'rewrite'}) {
					$proxy_rules =~ s/\r//;
					foreach (split ("\n",$proxy_rules)) {
						if (my $cond = rewrite($r, $_, \$uri)) {
							if ($cond eq "R") {
								$r->status(200);
								$r->headers_out->set('Location' => $uri);
								return Apache2::Const::REDIRECT;
							}
							if ($cond ne "P") {
								return $r->status;
							}
						}		
					}
				}

				if ($app->{'sso_forward_type'} eq "basic") {
					$r->headers_in->set('Authorization' => $session{post});
				}
				my $sth = $dbh->prepare("SELECT var, value, type FROM headers, app WHERE id_app='".$app->{id}."' AND app.id=id_app AND app.has_headers='t'");
				$sth->execute;
				while (my ($var, $value, $type) = $sth->fetchrow) {
					$value = $r->connection->remote_ip if ($type == REMOTE_ADDR);
					$value = $r->ssl_lookup($headers_vars{$type}) if (exists $headers_vars{$type});
					$r->headers_in->set($var => $value);
				}
				
				if (!$session{url}) {
				    my $sth = $dbh->prepare("SELECT url FROM app_lb WHERE id_app='".$app->{id}."' ORDER BY RANDOM() LIMIT 1");
				    $sth->execute;
				    if (my ($url) = $sth->fetchrow) {
					session(\%session) if (!$id);
					$session{url} = $url;
					$r->err_headers_out->add('Set-Cookie' => 'vulture_app='.$session{_session_id}.'; path=/');
				    }
				    else {
					$session{url} = $app->{url};
				    }
			    }
				$r->filename("proxy:".$session{url}.$uri);
				$r->handler('proxy-server');
				$r->proxyreq(1);
				return Apache2::Const::OK;
			}
			$r->headers_out->add('Set-Cookie' => 'vulture_app='.$session{_session_id}.'; path=/');
			my $request = HTTP::Request->new('POST', $app->{post_url}, undef, $session{post});
			$session{posted} = 1;
			$request->push_header('Content-Type' => 'application/x-www-form-urlencoded');
			my $ua = LWP::UserAgent->new;
			my $response = $ua->request($request); 
			if ($response->headers->header('Set-Cookie')) {
				foreach ($response->headers->header('Set-Cookie')) {
					$r->headers_out->add('Set-Cookie' => $_ . "; path=/");
				}
			}
		}
	}
	$r->status(Apache2::Const::REDIRECT);
	return Apache2::Const::OK;
}


1;
